/* ifile - intelligent mail filter for EXMH/MH
   Copyright (C) 1997  Jason Rennie <jr6b+@andrew.cmu.edu>

   irefile.c code Copyright (C) 1997 by Jason Rennie <jr6b+@andrew.cmu.edu>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program (see file 'COPYING'); if not, write to the Free
   Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA  02111-1307, USA.
   */

/*
  irefile.c

  C implementation of irefile.perl.

  Written by Jason Rennie, beginning Mar 30, 1997.

  This is a fairly straight translation from the perl code.  The only purpose
  at this point was to increase the efficiency of the program.  Some code was
  taken from the ifilter.c code.  The design of the code is very similar to
  that of the irefile.perl code.

  This program relies on the availability of the MH refile program.

  */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/stat.h>        // stat functions
#include <unistd.h>          // stat functions
#include <sys/types.h>       // directory functions
#include <dirent.h>          // directory functions
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <ctype.h>           // tolower function
#include "extendable_array.h"
#include "assoc_array_int.h"

/* rcvstore path --configure to fit your system. */
/* this is configured in the install.perl script.  Do not change manually! */
#define RCVSTORE "/usr/lib/mh/rcvstore"

/* properly defines CLOCKS_PER_SECONDS, if not already defined */
#ifndef CLOCKS_PER_SECOND
  #ifdef CLOCKS_PER_SEC
  #define CLOCKS_PER_SECOND CLOCKS_PER_SEC
  #else
    #ifdef CLK_TCK
    #define CLOCKS_PER_SECOND CLK_TCK
    #else
    #define CLOCKS_PER_SECOND 100
    #endif
  #endif
#endif

FILE *TMP;


/* Read a line, stripping the feedline '\n' */
/* void readline(char *s, int n, FILE *stream) { */
#define readline(s,n,stream) {\
    fgets(s, n, stream);\
    if(s[strlen(s)-1]=='\n') {\
	s[strlen(s)-1]='\0';\
    }\
}


/* Complain and die */
void ouch(char *msg) {
    fprintf(stderr, "%s: %s\n", msg, strerror(errno));
    fprintf(TMP, "%s: %s\n", msg, strerror(errno));
    exit(1);
}


/* prints information about program usage */
void print_usage()
{
  printf("Usage: irefile [msgs] [-src +folder] [-file file] [MH options] +folder\n");
  printf("       irefile -help\n\n");
  printf(" msgs         => text files to be refiled into FOLDER\n");
  printf(" -src +folder => looks for MSGS in specified folder\n");
  printf(" -file file   => disregards previous two options; uses 'file' as\n");
  printf("                 source message\n");
  printf(" +folder      => folder to direct messages into\n");
  printf(" MH options   => other command line options will be passed to\n");
  printf("                 MH refile, so all MH refile options are valid\n\n");
  exit(0);
}


/* Pass message to rcvstore */
void rcvstore(extendable_array lines, char *mbox) {
    int pfd[2];  /* Pipe file descriptors */
    char *mboxplus;
    int i;

    mboxplus=(char*)malloc(strlen(mbox)+2);
    strcpy(mboxplus,"+");
    strcat(mboxplus, mbox);

    fprintf(TMP, "Filed to %s\n", mbox);

    if(pipe(pfd)==-1) {
	ouch("Could not open pipe to rcvstore.");
    }
    switch(fork()) {
    case -1:   /* Error */
	ouch("Fork to rcvstore failed.");
	exit(1);
    case 0:    /* Child */
	if(close(pfd[1])==-1)   /* We don't need the writing end */
	    ouch("Close of writing end in child failed.");
	dup2(pfd[0], 0);  /* Make stdin refer to the inherited reading
			     end of the pipe. */
	execlp(RCVSTORE, RCVSTORE, mboxplus, NULL);
	ouch("execlp to rcvstore failed.\n");
	exit(1);
    default:   /* Parent */
	if(close(pfd[0])==-1)  /* We don't need the reading end */
	    ouch("Close of reading end in parent failed.");
	dup2(pfd[1], 1);  /* Make stdout refer to the writing end. */
	for(i=0; i<lines.elems; i++) {
	    printf("%s\n", EXT_ARRAY_GET(lines,char*,i));
	}
    }
}


/* returns 1 if /usr/bin/mh/refile.bak is executable, 0 otherwise */
int refile_exe()
{
  struct stat * tmp_file_stat = (struct stat *) malloc(sizeof(stat));

  if (!(stat("/usr/bin/mh/refile.bak", tmp_file_stat) == 0))
    {
      fprintf(stderr, "/usr/bin/mh/refile.bak does not stat!  Error: %s\n", strerror(errno));
      return 0;
    }

  // checks to make sure that the backup file is executable
  if (tmp_file_stat->st_mode && S_IXOTH == 0)
    {
      fprintf(stderr, "/usr/bin/mh/refile.bak is not executable!");
      return 0;
    }
  else
    {
      fprintf(stderr, "/usr/bin/mh/refile.bak is executable.\n");
      return 1;
    }
}

/* stops current process and runs MH version of refile
   returns -1 upon failure 
   */
int mh_refile(int argc, char **argv)
{
  /* The stat function was giving me hell, so I decided to eliminate the 
     check for exec priveleges.  If you get rid of the comments, you can
     try the stat code for yourself.  For some reason stat() would always
     return -1 for me.
     */

  char * mh_refile = (char *) malloc(strlen("/usr/bin/mh/refile.bak") + 1);
  char * arg = NULL;
  /*struct stat * tmp_file_stat = malloc(sizeof(stat)); */
  int i;
  int arg_size = 0;

  strcpy(mh_refile, "/usr/bin/mh/refile.bak");

  for (i=0; i < argc; i++) arg_size += strlen(argv[i]) + 1;
  arg = (char *) malloc(arg_size);

  strcpy(arg, mh_refile);
  fprintf(TMP, "%s", mh_refile);
  for (i=1; i < argc; i++)
    {
      strcat(arg, " ");
      strcat(arg, argv[i]);
      fprintf(TMP, " %s", argv[i]);
    }

  /*  if (stat(mh_refile, tmp_file_stat))
    {
      fprintf(TMP, "%s did not stat!\n", mh_refile);
      return -1;
    }
    */

  /* checks to make sure that the backup file is executable */
  /* if (tmp_file_stat->st_mode && S_IXOTH == 0)
    return -1;
  else
    { */
      fprintf(TMP, "\nUsing refile.bak to refile message: %s\n", arg);
      return system(arg);
      /* } */
}


/* Main program */
int main(int argc, char **argv) {
    /* To the extent that it was possible, the original variable names
       were kept to avoid confusion. */
    clock_t DMZ_start, DMZ_end;
    int i, j, k;
    char *home_dir = NULL;
    char *mh_profile = NULL;
    FILE *MH = NULL;
    char *mail_path = NULL;
    char tmpstr[BUFSIZ*8], *tmpstrptr = NULL;
    char *draft_folder = NULL;
    char *source_folder = NULL;
    char *source_file = NULL;
    char *dest_folder = NULL;
    extendable_array orig_message;
    char *acc_file = NULL;
    FILE *ACC = NULL;
    int filters, refiles;
    float accuracy = 0.0;
    int tot_words;
    extendable_array tv;
    char *data_file = NULL;
    FILE *DATA = NULL;
    int tvsize;
    assoc_array_int n, docs, index, age, nk, sum_nk;
    int tot_docs;
    int tmpint;
    extendable_array word;
    char *curword = NULL;
    int cat_num, freq;
    char tmpstr2[BUFSIZ];
    extendable_array lines;
    char nonalpha[256];
    char *wrd = NULL;
    char *cur_wrd = NULL;
    int word_count;
    char * tmp;
    char * msg = NULL;
    char * msg_path = NULL;
    char * source_folder_path = NULL;
    FILE * MSG = NULL;
    int skip_msg = 0;
    extendable_array f_tv;
    int val;
    char * cur_word = NULL;
    int weight;

    /* Get initial time */
    DMZ_start=clock();

    /* Open log file and write initial info */
    TMP=fopen("/tmp/irefile.info", "a");
    if(TMP==NULL)
      ouch("Not able to open .info file");

    fprintf(TMP, "irefile (C version) called (");
    for(i=0; i<argc; i++) {
	fprintf(TMP, "%s", argv[i]);
	if (i<argc-1) fprintf(TMP, " ");
    }
    fprintf(TMP, ")\n");

    /* Get home directory */
    home_dir=getenv("HOME");
    if(home_dir==NULL)
      {
	fclose(TMP);
	ouch("Fatal: HOME environment variable not defined!\n");
      }

    /* Find mail dir in .mh_profile */
    mh_profile = (char*) malloc(strlen(home_dir)+strlen("/.mh_profile")+1);
    strcpy(mh_profile, home_dir);
    strcat(mh_profile, "/.mh_profile");
    MH=fopen(mh_profile, "r");
    /* If .mh_profile is not found, set default Mail directory */
    if(MH==NULL)
      mail_path="Mail/";
    else 
      {
	while(!feof(MH)) 
	  {
	    readline(tmpstr, sizeof(tmpstr), MH);
	    tmpstrptr=strtok(tmpstr, " \t");
	    if(strcmp(tmpstrptr, "Path:")==0)
	      {
		tmpstrptr=strtok(NULL, " \t");
		mail_path=(char*)malloc(strlen(tmpstrptr)+2);
		strcpy(mail_path, tmpstrptr);
		strcat(mail_path, "/");
	      }
	    else if(strcmp(tmpstrptr, "draft-folder:")==0)
	      {
		tmpstrptr=strtok(NULL, " \t");
		draft_folder=(char*)malloc(strlen(tmpstrptr)+1);
		strcpy(draft_folder, tmpstrptr);
	      }
	  }
	fclose(MH);
      }

    EXT_ARRAY_INIT(orig_message, char *, 10);
    
    /* gets command line arguments */
    for (i=1; i < argc; i++)
      {
	if (strcmp(argv[i], "-src") == 0)
	  {
	    if (argv[i+1][0] == '+')
	      {
		source_folder = (char *) malloc(strlen(argv[i+1]));
		strcpy(source_folder, argv[i+1] + 1);
		i++;
	      }
	    else
	      {
		fprintf(TMP, "Error: folder name %s lacks preceeding \'+\' or is in wrong format.\n", source_folder);
		print_usage();
	      }
	  }
	else if (strcmp(argv[i], "-file") == 0)
	  {
	    source_file = (char *) malloc(strlen(argv[i+1])+1);
	    strcpy(source_file, argv[i+1]);
	    fprintf(TMP, "source file = %s\n", source_file);
	    i++;
	  }
	else if ((strcmp(argv[i], "--help") == 0) ||
		 (strcmp(argv[i], "-h") == 0))
	  print_usage();
	else if ((strcmp(argv[i], "-draft") == 0) ||	    
		 (strcmp(argv[i], "-nolink") == 0) ||
		 (strcmp(argv[i], "-preserve") == 0) ||
		 (strcmp(argv[i], "-nopreserve") == 0) ||
		 (strcmp(argv[i], "-rmmproc") == 0))
	  { }
	else if (strcmp(argv[i], "-link") == 0)
	  {
	    if (mh_refile(argc, argv) == -1)
	      ouch("Not able to call MH refile.\nDon't know what to do with -link option!\nNo refile action taken.\n");
	    else
	      exit(0);
	  }
	else if (argv[i][0] == '+')
	  {
	    dest_folder = (char *) malloc(strlen(argv[i]));
	    strcpy(dest_folder, argv[i]+1);
	  }
	else
	  {
	    tmp = (char *) malloc(strlen(argv[i])+1);
	    strcpy(tmp, argv[i]);
#ifdef DEBUG
	    fprintf(TMP, "message name = %s\n", tmp);
#endif
	    j = orig_message.last_elem + 1;
	    EXT_ARRAY_SET(orig_message, char *, j, tmp);
	    tmp = NULL;

#ifdef DEBUG
	    j = orig_message.last_elem;
	    printf("orig_message[%d] = %s\n", orig_message.last_elem,
		   EXT_ARRAY_GET(orig_message, char *, j));
#endif

	  }
      }

    if (dest_folder == NULL)
      {
	fprintf(TMP, "I need a destination folder.\n");
	fprintf(stderr, "I need a destination folder.\n");
	print_usage();
      }

    if (source_folder == NULL)
      {
	FILE * CONTEXT;
	char * context_file = malloc(strlen(home_dir) + strlen(mail_path) +
				     strlen("context") + 2);
	strcpy(context_file, home_dir);
	strcat(context_file, "/");
	strcat(context_file, mail_path);
	strcat(context_file, "context");
	
	if ((CONTEXT = fopen(context_file, "r")) == NULL)
	  {
	    sprintf(tmpstr,
		    "Couldn't open %s!  Don't know which folder to get messages from!\n", context_file);
	    ouch(tmpstr);
	  }

	readline(tmpstr, sizeof(tmpstr), CONTEXT);
	while (!feof(CONTEXT))
	  {
	    tmpstrptr = strtok(tmpstr, " \t");
	    if (strcmp(tmpstrptr, "Current-Folder:") == 0)
	      {
		tmpstrptr = strtok(NULL, " \t");
		source_folder = malloc(strlen(tmpstrptr)+1);
		strcpy(source_folder, tmpstrptr);
#ifdef DEBUG
		fprintf(TMP, "Current folder is %s\n", tmpstrptr);
		fprintf(stderr, "Current folder is %s\n", tmpstrptr);
#endif
	      }
	    readline(tmpstr, sizeof(tmpstr), CONTEXT);
	  }		     
      }

    if (orig_message.last_elem == -1)
      {
	FILE * MH_SEQ;
	char * mh_seq = malloc(strlen(home_dir) + strlen(mail_path) +
			       strlen(source_folder) +
			       strlen(".mh_sequences") + 3);
	strcpy(mh_seq, home_dir);
	strcat(mh_seq, "/");
	strcat(mh_seq, mail_path);
	strcat(mh_seq, source_folder);
	strcat(mh_seq, "/");
	strcat(mh_seq, ".mh_sequences");
	
	if ((MH_SEQ = fopen(mh_seq, "r")) == NULL)
	  {
	    sprintf(tmpstr,
		    "Couldn't open %s!  Can't determine current message!\n",
		    mh_seq);
	    ouch(tmpstr);
	  }

	readline(tmpstr, sizeof(tmpstr), MH_SEQ);
	while (!feof(MH_SEQ))
	  {
	    tmpstrptr = strtok(tmpstr, " \t");
	    if (strcmp(tmpstrptr, "cur:") == 0)
	      {
		tmpstrptr = strtok(NULL, " \t");
		tmp = (char *) malloc(strlen(tmpstrptr)+1);
		strcpy(tmp, tmpstrptr);
		EXT_ARRAY_SET(orig_message, char *, 0, tmp);
#ifdef DEBUG
		fprintf(TMP, "Current message is %s\n", tmp);
		fprintf(stderr, "Current message is %s\n", tmp);
#endif
		tmp = NULL;
	      }
	    readline(tmpstr, sizeof(tmpstr), MH_SEQ);
	  }
	free(mh_seq);
      }

    if (orig_message.last_elem == -1)
	ouch("No cur message.\n");

    if (source_file != NULL)
      {
	if (orig_message.last_elem >= 0)
	  {
	    fprintf(stderr, "Messages other than %s are being ignored.\n",
		   source_file);
	    fprintf(TMP, "Messages other than %s are being ignored.\n",
		    source_file);

	    /* no use for orig_message strings - free used memory */
	    for (i=0; i <= orig_message.last_elem; i++)
	      {
		free(EXT_ARRAY_GET(orig_message, char *, i));
		EXT_ARRAY_SET(orig_message, char *, i, NULL);
	      }
	    orig_message.elems = 0;
	    orig_message.last_elem = -1;
	  }

	if (source_folder != NULL)
	  {
	    fprintf(stderr, "\'-src\' command line option is being ignored\n");
	    fprintf(TMP, "\'-src\' command line option is being ignored\n");
	    free(source_folder);
	    source_folder = NULL;
	  }
	/* set 0th orig_message equal to source_file */
	tmp = malloc(strlen(source_file)+1);
	strcpy(tmp, source_file);
	EXT_ARRAY_SET(orig_message, char *, 0, tmp);
	tmp = NULL;
      }
	     
#ifdef DEBUG
    fprintf(TMP, "source folder = %s\n", source_folder);
    fprintf(TMP, "destination folder = %s\n", dest_folder);
    fprintf(TMP, "home directory = %s\n", home_dir);
#endif

    /* Update accuracies file */
    acc_file=(char*)malloc(strlen(home_dir)+strlen("/.idata_accuracy")+1);
    strcpy(acc_file, home_dir);
    strcat(acc_file, "/.idata_accuracy");
    filters=0;
    refiles=orig_message.elems;

    ACC=fopen(acc_file, "r");
    if(ACC!=NULL) 
      {
	readline(tmpstr, sizeof(tmpstr), ACC);
	tmpstrptr=strtok(tmpstr, " \t=");
	if(strcmp(tmpstrptr, "filters")==0) 
	  {
	    filters+=atoi(strtok(NULL, " \t="));
	    tmpstrptr=strtok(NULL, " \t=");
	    if(strcmp(tmpstrptr, "refiles")==0)
	      {
		tmpstrptr=strtok(NULL, " \t=");
		refiles+=atoi(tmpstrptr);
	      }
	  }
	accuracy=(((float)(filters-refiles)/(float)filters)*10000.0)/100.0;
	fprintf(TMP, "filters = %d  refiles = %d\n", filters, refiles);
	fprintf(TMP, "Accuracy = %.2f %%\n", accuracy);
	
	fclose(ACC);
    }
    
    ACC=fopen(acc_file, "w");
    if(ACC!=NULL)
      {
	/* reset accuracies file if unusual situation has come arisen */
	if (refiles > filters && refiles > 30)
	  {
	    refiles = 0;
	    filters = 0;
	    accuracy = 0.0;
	  }

	fprintf(ACC, "filters = %d  refiles = %d\n", filters,
		refiles);
	fprintf(ACC, "Accuracy = %.2f %%\n", accuracy);
	fclose(ACC);
    }
    else {
	fprintf(TMP, "Was not able to write accuracies file\n");
    }

    /* Now read the database */
    tot_words=0;

    EXT_ARRAY_INIT(tv,char*,EXT_ARRAY_CHUNK_SIZE);
    EXT_ARRAY_SET(tv,char*,0,NULL);
    tvsize=0;

    data_file=(char*)malloc(strlen(home_dir)+strlen("/.idata")+1);
    strcpy(data_file, home_dir);
    strcat(data_file, "/.idata");

    /* initialize variables to store info from database */
    assoc_array_int_init(n);
    assoc_array_int_init(docs);
    assoc_array_int_init(index);
    assoc_array_int_init(age);
    assoc_array_int_init(nk);
    assoc_array_int_init(sum_nk);
    EXT_ARRAY_INIT(word,char*,1000);

    if((DATA = fopen(data_file, "r")) != NULL) {

	/* Read the category names */
	readline(tmpstr, sizeof(tmpstr), DATA);
	tmpstrptr=strtok(tmpstr, " \t");
	while(tmpstrptr!=NULL) {
	    tvsize++;
	    EXT_ARRAY_SET(tv,char*,tvsize,strdup(tmpstrptr));
	    tmpstrptr=strtok(NULL, " \t");
	}

	/* Read total number of words for each category */
	readline(tmpstr, sizeof(tmpstr), DATA);
	i=1;
	tmpstrptr=strtok(tmpstr, " \t");
	while(tmpstrptr!=NULL) {
	    assoc_array_int_set(n, EXT_ARRAY_GET(tv,char*,i),
				atoi(tmpstrptr));
	    i++;
	    tmpstrptr=strtok(NULL, " \t");
	}
	if(i<=tvsize) {
	    ouch("Bad data file format - line #2");
	}

	/* Read number of documents for each category */
	readline(tmpstr, sizeof(tmpstr), DATA);
	tot_docs=0;
	i=1;
	tmpstrptr=strtok(tmpstr, " \t");
	while(tmpstrptr!=NULL) {
	    tmpint=atoi(tmpstrptr);
	    assoc_array_int_set(docs, EXT_ARRAY_GET(tv,char*,i),tmpint);
	    i++;
	    tot_docs+=tmpint;
	    tmpstrptr=strtok(NULL, " \t");
	}
	if(i<=tvsize) {
	    ouch("Bad data file format - line #3");
	}

	/* Read word frequencies */
	EXT_ARRAY_SET(word,char*,0,NULL);

	readline(tmpstr, sizeof(tmpstr), DATA);
	while(!feof(DATA)) {
	    tmpstrptr=strtok(tmpstr, " \t");
	    if(tmpstrptr==NULL)
		break;
	    tot_words++;
	    EXT_ARRAY_SET(word,char*,tot_words,strdup(tmpstrptr));
	    assoc_array_int_set(index, tmpstrptr, tot_words);
	    curword=tmpstrptr;

	    /* Read age of word */
	    tmpstrptr=strtok(NULL, " \t");
	    assoc_array_int_set(age, curword, atoi(tmpstrptr));

	    /* Read frequency per category cat_num:freq */
	    tmpstrptr=strtok(NULL, ":");
	    while(tmpstrptr!=NULL) {
		cat_num=atoi(tmpstrptr);
		tmpstrptr=strtok(NULL, " \t");
		freq=atoi(tmpstrptr);
		sprintf(tmpstr2, "%s,%s",
			EXT_ARRAY_GET(word,char*,tot_words),
			EXT_ARRAY_GET(tv,char*,cat_num));
		assoc_array_int_set(nk, tmpstr2, freq);
		/* sums frequencies per word for later usage */
		assoc_array_int_set(sum_nk, EXT_ARRAY_GET(word,char*,tot_words), assoc_array_int_get(sum_nk, EXT_ARRAY_GET(word,char*,tot_words)) + freq);

		tmpstrptr=strtok(NULL, ":");
	    }
	    readline(tmpstr, sizeof(tmpstr), DATA);
	}
    }

    /* Print out times, database stats */
    DMZ_end=clock();
    fprintf(stderr,"Read database in %.3f seconds.\n", 
	    ((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
    fprintf(TMP,"Read database in %.3f seconds.\n",
	    ((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
    fprintf(stderr, "%d words, %d categories read\n", tot_words, tvsize);
    fprintf(TMP, "%d words, %d categories read\n", tot_words, tvsize);

    fclose(DATA);

    DMZ_start=clock();

#ifdef DEBUG
    printf("home_dir = %s  mail_path = %s  source_folder = %s\n",
	   home_dir, mail_path, source_folder);
#endif

    source_folder_path = malloc(strlen(home_dir) + strlen(mail_path) +
				strlen(source_folder) + 3);
    strcpy(source_folder_path, home_dir);
    strcat(source_folder_path, "/");
    strcat(source_folder_path, mail_path);
    strcat(source_folder_path, source_folder);
    strcat(source_folder_path, "/");    

#ifdef DEBUG
    printf("source_folder_path = %s\n", source_folder_path);
#endif

    /* refiles messages one at a time */

    for (i=0; i <= orig_message.last_elem; i++)
      {
	skip_msg = 0;

#ifdef DEBUG
	printf("Current message is %d\n", i);
#endif
#ifdef DEBUG
	printf("orig_message[%d] = %s\n", i,
	       EXT_ARRAY_GET(orig_message, char *, i));
#endif
	
	if (source_file == NULL)
	  {
	    msg = malloc(strlen(EXT_ARRAY_GET(orig_message, char *, i))+1);
	    strcpy(msg, EXT_ARRAY_GET(orig_message, char *, i));

	    msg_path = malloc(strlen(msg) + strlen(source_folder_path) + 1);
	    strcpy(msg_path, source_folder_path);
	    strcat(msg_path, msg);
	  }
	else
	  {
	    msg_path = malloc(strlen(source_file)+1);
	    strcpy(msg_path, source_file);
	  }

	if (source_folder != NULL)
	  {
	    fprintf(TMP, "Opening %s\n", msg_path);
	    if ((MSG = fopen(msg_path, "r")) == NULL)
	      {
		fprintf(TMP, "Could not open %s\n", msg_path);
		skip_msg = 1;
	      }
	  }
	else
	  {
	    fprintf(TMP, "Opening %s\n", msg);
	    if ((MSG = fopen(msg, "r")) == NULL)
	      {
		fprintf(TMP, "Could not open %s\n", msg);
		skip_msg = 1;
	      }
	  }

free(msg);
	
	/* Initialize the non-alphabetic-character token separator */
	k=0;
	for(j=1; j<=255; j++) 
	  {
	    if(!((j>='a' && j<='z') || (j>='A' && j<='Z')))
	      nonalpha[k++]=j;
	  }
	nonalpha[k]='\0';
	
	EXT_ARRAY_INIT(lines,char*,EXT_ARRAY_CHUNK_SIZE);
	j=0;
	word_count=0;
	
	/* Read the message in an array for processing */
	readline(tmpstr, sizeof(tmpstr), MSG);
	while(!feof(MSG)) 
	  {
	    weight = 1;
	    k = lines.last_elem + 1;
	    EXT_ARRAY_SET(lines, char*, k, strdup(tmpstr));
	    j++;
	    /* Split line in words and store them in another array */
	    tmpstrptr = strtok(tmpstr, nonalpha);
	    while (tmpstrptr != NULL)
	      {
		wrd = tmpstrptr;
		/* Skip tiny words */
		if (strlen(wrd) > 2) 
		  {
		    cur_wrd = wrd;
		    /* Convert to lower case */
		    while((*wrd) != '\0') 
		      {
			(*wrd) = tolower(*wrd);
			wrd++;
		      }

		    /* Update vocab list and word statistics */
		    if (assoc_array_int_get(index, cur_wrd) == 0)
		      {
#ifdef DEBUG
			printf("Creating new entry for %s\n", cur_wrd);
#endif
			tmp = malloc(strlen(cur_wrd)+1);
			strcpy(tmp, cur_wrd);
			k = word.last_elem+1;
			EXT_ARRAY_SET(word, char *, k, tmp);
			assoc_array_int_set(index, cur_wrd, k);

			tmp = malloc(strlen(cur_wrd) + strlen(dest_folder) + 2);
			sprintf(tmp, "%s,%s", cur_wrd, dest_folder);
			assoc_array_int_set(nk, tmp, weight + 
					    assoc_array_int_get(nk, tmp));
			assoc_array_int_set(sum_nk, cur_wrd, assoc_array_int_get(sum_nk, cur_wrd) + weight);

			free(tmp);
		      }
		    else
		      {	
			tmp = malloc(strlen(cur_wrd) +
				     strlen(dest_folder) + 2);
			sprintf(tmp, "%s,%s", cur_wrd, dest_folder);
			assoc_array_int_set(nk, tmp, weight + 
					    assoc_array_int_get(nk, tmp));
			assoc_array_int_set(sum_nk, cur_wrd, assoc_array_int_get(sum_nk, cur_wrd) + weight);
			free(tmp);
		      }
		    assoc_array_int_set(n, dest_folder, weight +
					assoc_array_int_get(n, dest_folder));
#ifdef VERBOSE
		    tmp = malloc(strlen(cur_wrd) +
				 strlen(dest_folder) + 2);
		    sprintf(tmp, "%s,%s", cur_wrd, dest_folder);
		    printf("word freq (nk) of %s is %d\n", tmp,
			   assoc_array_int_get(nk, tmp));
		    free(tmp);
#endif
		  }
		tmpstrptr = strtok(NULL, nonalpha);
	      }
	    readline(tmpstr, sizeof(tmpstr), MSG);
	  }

	fclose(MSG);

	assoc_array_int_set(docs, dest_folder, 1 +
			    assoc_array_int_get(docs, dest_folder));	
	for (j=1; j <= word.last_elem; j++)
	  assoc_array_int_set(age, EXT_ARRAY_GET(word, char *, j), 1 +
			      assoc_array_int_get(age, EXT_ARRAY_GET(word, char *, j)));

	/* print out timings */
	DMZ_end=clock();
	fprintf(stderr,"Read %d message in %.3f seconds.\n",
		orig_message.elems,
		((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
	fprintf(TMP,"Read %d message in %.3f seconds.\n", orig_message.elems,
		((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
      }
    
    /* Reopen and write the learned information */

    DMZ_start = clock();

    data_file=(char*)malloc(strlen(home_dir)+strlen("/.idata")+1);
    strcpy(data_file, home_dir);
    strcat(data_file, "/.idata");

    EXT_ARRAY_INIT(f_tv, char *, 50);

    if((DATA = fopen(data_file, "w")) != NULL) {

      /* Reads current folder list from "~/[mail_path]/.folders", keeps the
         names in memory and writes them to disk as the first line of .idata */

      char * folders_file = NULL;
      FILE * FOLDERS = NULL;

      folders_file = (char *) malloc(strlen(home_dir) + strlen(mail_path) +
				     strlen("/.folders") + 1);
      strcpy(folders_file, home_dir);
      strcat(folders_file, "/");
      strcat(folders_file, mail_path);
      strcat(folders_file, ".folders");
      
      EXT_ARRAY_INIT(f_tv, char *, 50);

      if ((FOLDERS = fopen(folders_file, "r")) != NULL)
	{
	  for (i=1; fgets(tmpstr, sizeof(tmpstr), FOLDERS); i++)
	    {
	      if (tmpstr[strlen(tmpstr)-1] == '\n')
		tmpstr[strlen(tmpstr)-1] = '\0';
	      EXT_ARRAY_SET(f_tv, char *, i, malloc(strlen(tmpstr)+1));
	      strcpy(EXT_ARRAY_GET(f_tv, char *, i), tmpstr);
	      fputs(tmpstr, DATA);
	      fputc(' ', DATA);
	    }
	}
      else
	ouch("Could not locate .folders file");

      fputc('\n', DATA);

      /* Writes to .idata the total number of words observed for each of the
         folders */

      for (i=1; i <= f_tv.last_elem; i++)
	{
	  val = assoc_array_int_get(n, EXT_ARRAY_GET(f_tv, char *, i));
	  if (val > 0)
	    fprintf(DATA, "%d ", val);
	  else
	    fprintf(DATA, "0 ");
	}
      
      fputc('\n', DATA);

      /* Writes to .idata the total number of documents observed for each of
	 the folders */

      for (i=1; i <= f_tv.last_elem; i++)
	{
	  val = assoc_array_int_get(docs, EXT_ARRAY_GET(f_tv, char *, i));
	  if (val > 0)
	    fprintf(DATA, "%d ", val);
	  else
	    fprintf(DATA, "0 ");
	}
      
      fputc('\n', DATA);

      /* sum_nk is never created properly */
      

      /* For each word in the vocabulary, writes frequency by folder if the
	 word is frequent enough to warrant recording */

      for (i=1; i <= word.last_elem; i++)
	{
	  cur_word = EXT_ARRAY_GET(word, char *, i);
	  /* if the word is infrequent, discard it (don't print it to DATA) */
	  if ((assoc_array_int_get(sum_nk, cur_word) > 3) ||
	      (assoc_array_int_get(age, cur_word) < 16))
	    {
	      fprintf(DATA, "%s %d ", cur_word,
		      assoc_array_int_get(age, cur_word));

	      for (j=1; j <= f_tv.last_elem; j++)
		{
		  sprintf(tmpstr, "%s,%s",
			  EXT_ARRAY_GET(word, char *, i),
			  EXT_ARRAY_GET(f_tv, char *, j));
		  val = assoc_array_int_get(nk, tmpstr);
#ifdef VERBOSE
		  printf("word freq (nk) of %s is %d\n", tmpstr, val);
#endif
		  if (val > 0)
		    fprintf(DATA, "%d:%d ", j, val);
		}
	      fputc('\n', DATA);
	    }
	}
    }
    fclose(DATA);
    
    DMZ_end=clock();
    fprintf(stderr,"Wrote database in %.3f seconds.\n", 
	    ((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
    fprintf(TMP,"Wrote database in %.3f seconds.\n",
	    ((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
    fprintf(stderr, "%d words, %d categories written\n", tot_words, tvsize);
    fprintf(TMP, "%d words, %d categories written\n", tot_words, tvsize);

    /* if refile.bak is executable, then use it to refile all the messages */
    if (refile_exe())
      mh_refile(argc, argv);
    
    exit(0);
}
